function result = run_exp(data, lambda_list, k_list, repeat_list, rho_list)
	if nargin < 2
		lambda_list = [ 1, 2, 3, 4];
	end
	if nargin < 3
		k_list = [4 5 6];
	end
	if nargin < 4
		repeat_list = [1 2];
	end
	if nargin < 5
		rho_list = [0 1 3 5];
	end

	result = struct();
	for i=1:length(lambda_list)
		lambda = 10^-lambda_list(i);
		for j=1:length(k_list)
			field = sprintf('l%dk%d', lambda_list(i), k_list(j));
			fprintf('====> start %s\n', field);
			k = 2^k_list(j);
			res = run_one_new(data, lambda, k, repeat_list, rho_list);
			result = setfield(result, field, res);
		end
	end
end


function result = run_one_new(data, lambda, k, repeat_list, rho_list)
	%data = gen_one_class(data);
	result = struct();

	% sampled schemes
	schemes = {'user','item_f','item_w','item_s','uniform'};
	for i=1:length(schemes)
		scheme = schemes{i};
		options = sprintf('-s 1 -t 10 -k %d -l %g -b 1 -n 8', k, lambda);
		for i = 1:length(repeat_list)
			m = sample_trainer(data.R, data.dim, repeat_list(i), scheme, options);
			ret = pmf_predict_ranking(data.Rt, m.W, m.H, 10, data.R);
			r.ndcg = ret.ndcg; r.map = ret.map; r.auc = ret.auc; r.hlu = ret.hlu;
			r.pos_rank = ret.pos_rank;
			field = sprintf('%s%d',scheme,repeat_list(i));
			result = setfield(result, field, r);
			fprintf('run %s: %g %g %g\n', field, r.ndcg(10), r.map, r.auc);
			disp(r);
		end
	end

	% full pu
	scheme = 'pu'; 
	for i = 1:length(rho_list)
		options = sprintf('-s 11 -t 10 -k %d -l %g -r %g -b 0 -n 8', k, lambda, 2^-rho_list(i));
		%m = sample_trainer(data.R, data.dim, 0, 'pu', options);
		m = pmf_train(data.R, [], options);
		ret = pmf_predict_ranking(data.Rt, m.W, m.H, 10, data.R);
		r.ndcg = ret.ndcg; r.map = ret.map; r.auc = ret.auc; r.hlu = ret.hlu;
		r.pos_rank = ret.pos_rank;
		field = sprintf('%s%d',scheme,rho_list(i));
		result = setfield(result, field, r);
		fprintf('run %s: %g %g %g\n', field, r.ndcg(10), r.map, r.auc);
		disp(r);
	end
end
